"""
#######################################################################
#
# Licensed Materials - Property of Pipelinefx L.L.C.
#
#
# (C) COPYRIGHT Pipelinefx Limited Liability Corporation.
#  All Rights Reserved.
#                               
#  US Government Users Restricted Rights - Use, duplication or
#  disclosure restricted by GSA ADP Schedule Contract with 
#  Pipelinefx L.L.C.
#
# QubeView Host Model
#
# Name: qbUIHost.py
#
# $DateTime: 2013/06/12 14:01:37 $
# $Revision: #1 $
# $Change: 11403 $
# $Id: //depot/main/qubeMobileServer/qubeMobile/hold_model/qbUIHost.py#1 $
# $File: //depot/main/qubeMobileServer/qubeMobile/hold_model/qbUIHost.py $
#
# Written by: Brian Knudson / PFX
#             briank@pipelinefx.com
#             2011-10-05
#
# Description:
#  UIHost class - Extends qb.Host with more ui-friendly methods.
#
#
#######################################################################
"""

import os, sys, re
import time
import datetime
import logging

try:
    import qb
except ImportError:
    api_path = "/Applications/pfx/qube/api/python/"
    if api_path not in sys.path:
        sys.path.insert(0,api_path)
    import qb


"""
Qb UI Host Base Class
"""
    

class QbUIHost(qb.Host):
    """
    A host model.
    """
    def __init__(self,host,job_list=None,verbosity=0):
        """
        Takes a required host (qb.Host) and an optional verbosity level to construct
        """
        super(QbUIHost,self).__init__(host)
        self._VERBOSITY = verbosity

        # store [address of] the job list so hosts can have access to it.
        self.__job_list = job_list

        # this is not strictly necessary, though it maintains some similarity between jobs & hosts
        self.qb_hostinfo = qb.hostinfo

        try:
            self.properties_dict = dict([raw_prop.split('=') for raw_prop in self['properties'].split(',')])
        except ValueError:
            self.properties_dict = self.__slowProps('properties')
        try:
            self.resources_dict  = dict([raw_res.split('=') for raw_res in self['resources'].split(',')])
        except ValueError:
            self.resources_dict = self.__slowProps('resources')

        self.jobtypes = sorted(self.get('jobtypes',[]).split(','))

        self.__has_detail = False;
        self.__config = None
        self.__subjobs = []
        self.__jobs = []
        self.__frames = []


    @property
    def frames(self):
        """
        If the list of frames doesn't exist, it populates the list.
        Otherwise, it returns the list
        """
        if not self.__frames:
            self.__updateFrames()
        return self.__frames

    @property
    def subjobs(self):
        """
        If the list of subjobs doesn't exist, it populates the list.
        Otherwise, it returns the list
        """
        if not self.__subjobs:
            self.__updateSubjobs()
        return self.__subjobs


    @property
    def jobs(self):
        """
        If the list of subjobs doesn't exist, it populates the list.
        Otherwise, it returns the list
        """
        if not self.__jobs:
            self.__updateJobs()
        return self.__jobs


    @property
    def state(self):
        """
        The 'state' property requires logic, so... make it a property
        """
        if self.resources_dict.get('host.processors') == '0/0' and self.get('state') == 'active':
            return 'locked'
        return self.get('state')


    def __slowProps(self,ptype):
        """
        Walk through properties & assign one at a time - slower, but more reliable
        """
        raw = {True:self['properties'], False:self['resources']}[ptype == 'properties']

        ret = {}

        for item in raw.split(','):
            try:
                ret.update(dict([item.split('=')]))
            except ValueError:
                ret[item] = ''

        return ret


    def __findStringInList(self,s,l):
        """
        Given a string and a list, this returns the first item in the list that matches the string
        """
        try:
            return l[map(lambda x: x.find(s) >= 0,l).index(True)]
        except ValueError:
            return None


    def __getMemory(self):
        """
        helper function
        NOTE: Not currently used - check properties_dict or resources_dict.
        Left in place for curiosity's sake
        """
        return self.__findStringInList("host.memory",self['resources'].split(',')).split('=')[-1]


    def __getRunningJobIds(self):
        """
        helper function to return a list of running jobs on this host
        """
        return [sj['pid'] for sj in self.get('subjobs',[])]


    def __getRunningSubjobIds(self):
        """
        helper function to return a list of running jobs on this host
        """
        return ["%s.%s" % (sj['pid'],sj['id']) for sj in self.get('subjobs',[])]


    def __getRunningUsers(self):
        """
        return a list of users whos jobs are running on this host
        """
        seen = set()
        unique_ids = [spid for spid in [sj['pid'] for sj in self.get('subjobs',[])] if spid not in seen and not seen.add(spid)]

        users = []
        for jid in unique_ids:
            jinfo = qb.jobinfo(id=jid)
            if not jinfo: continue
            user = jinfo[0]['user']
            if user not in users:
                users.append(user)
        return users


    def updateDetail(self, force=False):
        """
        Fully populate this host with options info, namely subjobs.
        Returns itself for faster access to data, but the return value is not
        strictly necessary as the object is now updated & all of the data in
        the return is also available in the object
        """
        logging.info("Updating detail for host '%s'" % self.get('name',self.get('address')))
        if self.__has_detail and not force:
            logging.debug("Skipping host detail because it's cached")
            return self
        
        # if self._VERBOSITY > 0:
        logging.info("updating host detail for host '%s'" % self['macaddress'])
        super(QbUIHost,self).__init__(self.qb_hostinfo(filters={'macaddress':self['macaddress']},subjobs=True)[0])
        self.getConfig(force_update=True)


        self.__updateJobs(force=True)
        self.__updateSubjobs(force=True)
        self.__updateFrames(force=True)
        self.__has_detail = True
        
        return self


    def __updateFrames(self,force=False):
        """
        Update the list of frames that this host is running at this moment
        The list of frames is returned for conveniece, though one could also access them using the
        property 'frames'
        """
        if self.__frames and not force:
            logging.debug("Returning cached frames for host '%s'" % self.get('name',''))
            return self.__frames()

        logging.debug("Updating frames for host '%s'" % self.get('name',''))
        subjob_ids = ["%s.%s" % (sj['pid'],sj['id']) for sj in self.get('subjobs',[])]

        # if there are no subjobs, there are no frames
        if not subjob_ids:
            return []

        # update our jobs if we need to:
        if force or not self.__jobs:
            self.__updateJobs(force=force)

        self.__frames = []
        for job in self.__jobs:
            self.__frames += [frame for frame in job.frames if "%s.%s" % (frame['pid'],frame['subid']) in subjob_ids if frame.get('status','') == 'running']

        return self.__frames
        
        

    def __updateSubjobs(self,force=False):
        """
        Update the list of subjobs that this host is running at this moment
        The list of subjobs is returned for conveniece, though one could also access them using the
        property 'subjobs'
        """
        if self.__subjobs and not force:
            logging.debug("Returning cached subjobs for host '%s'" % self.get('name',''))
            return self.__subjobs
        
        logging.debug("Updating subjobs for host '%s'" % self.get('name',''))
        subjob_ids = ["%s.%s" % (sj['pid'],sj['id']) for sj in self.get('subjobs',[])]
        if not subjob_ids:
            return []

        # update our jobs if we need to:
        if force or not self.__jobs:
            self.__updateJobs(force)

        self.__subjobs = []
        for job in self.__jobs:
            self.__subjobs += [sj for sj in job.subjobs if "%s.%s" % (sj['pid'],sj['id']) in subjob_ids if sj.get('status','') == 'running']

        return self.__subjobs


    def __updateJobs(self,force=False):
        """
        Update the list of jobs that this host is running at this moment.
        The list of jobs is returned for conveniece, though one could also access them using the
        property 'jobs'
        """
        if self.__jobs and not force:
            logging.debug("Returning cached jobs for host '%s'" % self.get('name',''))
            return self.__jobs

        logging.debug("Updating jobs for host '%s'" % self['name'])
        self.__jobs = []
        for jid in set([sj['pid'] for sj in self.get('subjobs',[])]):
            try:
                if force:
                    logging.debug("Host '%s' requesting an update of its running jobs" % self['name'])
                    self.__job_list[jid].updateDetail()
                self.__jobs.append(self.__job_list[jid])
            except Exception,e:
                logging.debug("Host could not get job %d.  Skipping: %s" % (jid,e))
                continue

        return self.__jobs
    
        

    def hasDetail(self):
        """
        Accessor function for the private 'has_detail'
        returns the val of the var
        """
        return self.__has_detail


    def getHostDetails(self):
        """
        Returns a dictionary of basic properties for this host as a dict.
        Basic props are: name, state, cluster, restrictions, groups, running jobs/users,
         mac/address, description, flags, last update
        Return keys are numerated to perserve order (maybe an ordered dict would be better) 
        """

        # Note: - this is a little too "controller" for a model.
        
        ret = {"01.Name":self['name'],
               "02.State":self.state,
               "03.Cluster":self['cluster'],
               "04.Groups":self['groups'],
               "06.Restrictions":"TBD",
               "07.Running Jobs":', '.join(map(str,set(self.__getRunningJobIds()))),
               "08.Running Users":', '.join(map(str,self.__getRunningUsers())),
               "09.IP Address":self['address'],
               "10.MAC Address":self['macaddress'],
               "11.Description":"TBD",
               "12.Flags":self['flagsstring'],
               "13.Last Updated":time.ctime(self['lastupdate'])
               }
        return ret


    def getConfig(self,force_update=False):
        """
        returns the current worker configuration as a dictionary.
        """
        if not self.__config or force_update:
            if self['state'] not in ('active','panic'):
                logging.info("Cannot connect to worker '%s' to get config.  Skipping." % self.get('name',self.get('address')))
            else:
                logging.info("Fetching worker config for '%s'" % self.get('name'))
                self.__config = qb.workerconfig(self['name'])
        return self.__config


    def getAttribute(self,att):
        """
        returns the attribute from this instance, given an attribute or human
        readable name.
        """

        att = att.lower()
        if att == 'worker mode':
            return self.properties_dict.get('host.worker_mode')
        elif att == 'processor speed':
            return self.properties_dict.get('host.processor_speed')
        elif att == 'architecture':
            return self.properties_dict.get('host.architecture')
        elif att == 'proxy mode':
            return self.properties_dict.get('host.proxy_mode')
        elif att == 'os':
            return self.properties_dict.get('host.os')
        elif att == 'build':
            return self.properties_dict.get('host.qube_build')
        elif att == 'kernel':
            return self.properties_dict.get('host.kernel')
        elif att == 'cpus':
            return self.properties_dict.get('host.cpus')
        elif att == 'version':
            return self.properties_dict.get('host.qube_version')
        elif att == 'cores':
            return self.resources_dict.get('host.cores')
        elif att == 'processors' or att == 'slots':
            return self.resources_dict.get('host.processors')
        elif att == 'memory':
            return self.resources_dict.get('host.memory')
        elif att == 'swap':
            return self.resources_dict.get('host.swap')
        elif att == 'running jobs':
            return ', '.join(map(str,self.__getRunningJobIds()))
        elif att == 'running users':
            return ', '.join(map(str,self.__getRunningUsers()))
        elif att == 'state':
            return self.state
        elif self.has_key(att):
            return self[att]
        else:
            return att
            
        
