"""
 Licensed Materials - Property of Pipelinefx L.L.C.


 (C) COPYRIGHT Pipelinefx Limited Liability Corporation.
  All Rights Reserved.
                               
  US Government Users Restricted Rights - Use, duplication or
  disclosure restricted by GSA ADP Schedule Contract with 
  Pipelinefx L.L.C.

 QubeView Frame module

 Name: qbUIFrame.py

 $DateTime: 2013/07/08 14:29:16 $
 $Revision: #3 $
 $Change: 11473 $
 $File: //depot/main/qubeMobileServer/qubeMobile/model/qbUIFrame.py $

 Written by: Brian Knudson / PFX
             briank@pipelinefx.com
             2011-10-01

 Description:
  Frame module, subclassed off of qb.Work with some extra functions to
  access logs using local cache
"""

import os
import sys
import re
import time
import datetime
import mmap
import logging
import tempfile

try:
    import qb
except ImportError:
    api_path = "/Applications/pfx/qube/api/python/"
    if api_path not in sys.path:
        sys.path.insert(0,api_path)
    import qb

from qbUIBase import QbUIBase

class QbUIFrame(qb.Work,QbUIBase):
    """
    A frame model.
    Construct with:
     frame: a qb.Work object
     log_list: a qbUILogList
    Note: Both outfp & errfp can be None, but they will never be updated in that case
    """
    attribute_map = [{"display":"Started",   "attribute":"getReadableTimeStart()",    "name":"started"},
                     {"display":"Completed", "attribute":"getReadableTimeComplete()", "name":"completed"},
                     {"display":"Elapsed",   "attribute":"getReadableElapsed()",      "name":"elapsed"},
                     {"display":"Output",    "attribute":"getOutput()",               "name":"output"},
                     {"display":"CPU-minutes","attribute":"getTimeCumulative()",      "name":"time_cumulative"},
                     {"display":"id"},
                     {"display":"pid"},
                     {"display":"subid"},
                     {"display":"host"},
                     {"display":"name"},
                     {"display":"status"},
                     {"display":"address"},
                     {"display":"data"},
                     {"display":"package"},
                     {"display":"result"},
                     {"display":"count"},
                     {"display":"retry"},
                     {"display":"retrydelay"},
                     {"display":"resultpackage"},
                     {"display":"lastupdate"},
                     {"display":"timesubmit"},
                     {"display":"timestart"},
                     {"display":"timecomplete"},
                     ]

    def __init__(self,frame={},log_list={}):
        super(QbUIFrame,self).__init__(frame)
        self.__log_list = log_list

        self.valid_time_frame_states = ("complete","running","failed")

    def getLogPaths(self):
        """
        Returns the paths to the temporary files used for this frame's logs
        return is (stdout_path,stderr_path)
        """
        return self.__log_list[self['pid']].getSubjobLogPath(self['subid'])

    def getLogs(self,update=False):
        """
        Return the log for this one frame
        return val is (stdout,stderr)
        """
        logging.debug("Fetching logs for frame '%s:%s'." % (self['pid'],self['id']))
        return self.__log_list[self['pid']].getFrameLogs(self['id']-1) # -1 to make up for 1 based array in the qb API

    def getTimeCumulative(self):
        try:
            return str(datetime.timedelta(seconds=self.get('timecumulative','--')))
        except TypeError:
            return '--'

    def getReadableTimeStart(self):
        try:
            return { True:time.strftime("%X",time.localtime(self["timestart"])), False:"--" }[ self["timestart"] > qb.QB_TIME_EPOCH_OFFSET ]
        except KeyError:
            return "--"

    def getReadableTimeComplete(self):
        try:
            return { True:time.strftime("%X",time.localtime(self["timecomplete"])), False:"--" }[ self["timecomplete"] > qb.QB_TIME_EPOCH_OFFSET ]
        except KeyError:
            return "--"

    def getReadableElapsed(self):
        """
        returns the time this frame has been running in a human
        readable string.
        """
        if not self['status'] in self.valid_time_frame_states:
            return "--"
        
        now = int(time.time())
        start_epoch    = {True:self['timestart'],    False:now}[self['timestart'] > qb.QB_TIME_EPOCH_OFFSET ]
        complete_epoch = {True:self['timecomplete'], False:now}[self['timecomplete'] > qb.QB_TIME_EPOCH_OFFSET ]
        elapsed_epoch = complete_epoch - start_epoch

        return str(datetime.timedelta(seconds=elapsed_epoch))

    def getOutput(self):
        rp = self.get("resultpackage",{})
        if rp:
            return rp.get("outputPaths")
        return None
        

