"""
 Licensed Materials - Property of Pipelinefx L.L.C.


 (C) COPYRIGHT Pipelinefx Limited Liability Corporation.
  All Rights Reserved.
                               
  US Government Users Restricted Rights - Use, duplication or
  disclosure restricted by GSA ADP Schedule Contract with 
  Pipelinefx L.L.C.

 $DateTime: 2013/07/08 14:29:16 $
 $Revision: #4 $
 $Change: 11473 $
 $File: //depot/main/qubeMobileServer/qubeMobile/model/qbUIHost.py $

 Written by: Brian Knudson / PFX
             briank@pipelinefx.com
             2011-10-05

 Description:
  UIHost class - Extends qb.Host with more ui-friendly methods.

"""

import os, sys, re
import time
import datetime
import logging
import traceback

try:
    import qb
except ImportError:
    for api_path in ("/Applications/pfx/qube/api/python/",
                     "/usr/local/pfx/qube/api/python/",
                     "C:\\Program Files\\pfx\\qube\\api\\python",
                     "C:\\Program Files (x86)\\pfx\\qube\\api\\python"):
        if api_path not in sys.path and os.path.exists(api_path):
            sys.path.insert(0,api_path)
            break
    import qb


from qbUIBase import QbUIBase

class QbUIHost(qb.Host,QbUIBase):
    """
    A host model.
    """
    attribute_map = [{"display":"worker mode",    "attribute":"properties_dict:host.worker_mode",     "name":"worker_mode"},
                     {"display":"processor speed","attribute":"properties_dict:host.processor_speed", "name":"processor_speed"},
                     {"display":"architecture",   "attribute":"properties_dict:host.architecture",    "name":"architecture"},
                     {"display":"proxy mode",     "attribute":"properties_dict:host.proxy_mode",      "name":"proxy_mode"},
                     {"display":"os",             "attribute":"properties_dict:host.os",              "name":"os"},
                     {"display":"build",          "attribute":"properties_dict:host.qube_build",      "name":"build"},
                     {"display":"kernel",         "attribute":"properties_dict:host.kernel",          "name":"kernel"},
                     {"display":"cpus",           "attribute":"properties_dict:host.cpus",            "name":"cpus"},
                     {"display":"qube version",   "attribute":"properties_dict:host.qube_version",    "name":"qube_version"},
                     {"display":"cores",          "attribute":"resources_dict:host.cores",            "name":"cores"},
                     {"display":"processors",     "attribute":"resources_dict:host.processors",       "name":"processors"},
                     {"display":"slots",          "attribute":"resources_dict:host.processors",       "name":"slots"},
                     {"display":"memory",         "attribute":"resources_dict:host.memory",           "name":"memory"},
                     {"display":"swap",           "attribute":"resources_dict:host.swap",             "name":"swap"},
                     {"display":"running jobs",   "attribute":"getRunningJobIds()",                   "name":"running_jobs"},
                     {"display":"running users",  "attribute":"getRunningUsers()",                    "name":"running_users"},
                     {"display":"state",          "attribute":"state()",                              "name":"state"},
                     {"display":"name"},
                     {"display":"macaddress"},
                     {"display":"cluster"},
                     {"display":"resources"},
                     {"display":"restrictions"},
                     {"display":"properties"},
                     {"display":"stats"},
                     {"display":"groups"},
                     {"display":"address"},
                     {"display":"subjobs"},
                     {"display":"jobtypes"},
                     ]



    def __init__(self,host={},job_list=None):
        """
        Construct with a qb.Host
        """
        super(QbUIHost,self).__init__(host)

        # store [address of] the job list so hosts can have access to it.
        self.__job_list = job_list

        # this is not strictly necessary, though it maintains some similarity between jobs & hosts
        self.qb_hostinfo = qb.hostinfo

        try:
            self.properties_dict = dict([raw_prop.split('=') for raw_prop in self['properties'].split(',')])
        except ValueError:
            self.properties_dict = self.__slowProps('properties')
        except KeyError:
            self.properties_dict = {}
        try:
            self.resources_dict  = dict([raw_res.split('=') for raw_res in self['resources'].split(',')])
        except ValueError:
            self.resources_dict = self.__slowProps('resources')
        except KeyError:
            self.resources_dict = {}

        try:
            self.jobtypes_list = sorted(self['jobtypes'].split(','))
        except KeyError:
            self.jobtypes_list = []

        self.__has_detail = False;
        self.__config = None
        self.__subjobs = []
        self.__jobs = []
        self.__frames = []


    @property
    def frames(self):
        """
        If the list of frames doesn't exist, it populates the list.
        Otherwise, it returns the list
        """
        if not self.__frames:
            self.__updateFrames()
        return self.__frames

    # Stopped here - this isn't working for some reason.
    @property
    def subjobs(self):
        """
        If the list of subjobs doesn't exist, it populates the list.
        Otherwise, it returns the list
        """
        if not self.__subjobs:
            self.__updateSubjobs()
        return self.__subjobs


    @property
    def jobs(self):
        """
        If the list of subjobs doesn't exist, it populates the list.
        Otherwise, it returns the list
        """
        if not self.__jobs:
            self.__updateJobs()
        return self.__jobs


    def state(self):
        """
        The 'state' property requires logic, so... make it a property
        """
        # if host.proc is x/0 & it's active, it's actually locked.
        if self.resources_dict.get('host.processors','').find('/0') >= 0 and self.get('state') == 'active':
            return 'locked'
        if self.get('state') == 'active' and len(self.subjobs) == 0:
            return 'idle'
        return self.get('state')


    def getRunningJobIds(self):
        """
        helper function to return a list of running jobs on this host
        """
        return [sj['pid'] for sj in self.get('subjobs',[])]


    def getRunningSubjobIds(self):
        """
        helper function to return a list of running jobs on this host
        """
        return ["%s.%s" % (sj['pid'],sj['id']) for sj in self.get('subjobs',[])]


    def getRunningUsers(self):
        """
        return a list of users whos jobs are running on this host
        """
        seen = set()
        unique_ids = [spid for spid in [sj['pid'] for sj in self.get('subjobs',[])] if spid not in seen and not seen.add(spid)]

        users = []
        for jid in unique_ids:
            jinfo = qb.jobinfo(id=jid)
            if not jinfo: continue
            user = jinfo[0]['user']
            if user not in users:
                users.append(user)
        return users


    def updateDetail(self, force=False):
        """
        Fully populate this host with options info, namely subjobs.
        Returns itself for faster access to data, but the return value is not
        strictly necessary as the object is now updated & all of the data in
        the return is also available in the object
        """
        logging.info("Updating detail for host '%s'" % self.get('name',self.get('address')))
        if self.__has_detail and not force:
            logging.debug("Skipping host detail because it's cached")
            return self
        
        logging.info("updating host detail for host '%s'" % self.get('macaddress','UNKNOWN'))
        try:
            super(QbUIHost,self).__init__(self.qb_hostinfo(filters={'macaddress':self['macaddress']},subjobs=True)[0])
        except KeyError:
            logging.debug("Cannot update host '%s'" % self.get('macaddress','UNKNOWN'))
            return
        self.getConfig(force_update=True)


        self.__updateJobs(force=True)
        self.__updateSubjobs(force=True)
        self.__updateFrames(force=True)
        self.__updateResources()
        if not self.properties_dict:
            self.__updateProperties()
        if not self.jobtypes_list:
            try:
                self.jobtypes_list = sorted(self['jobtypes'].split(','))
            except KeyError:
                pass
        self.__has_detail = True

        return self

    def __slowProps(self,ptype):
        """
        Walk through properties & assign one at a time - slower, but more reliable
        """
        raw = {True:self.get('properties',''), False:self.get('resources','')}[ptype == 'properties']

        ret = {}

        for item in raw.split(','):
            try:
                ret.update(dict([item.split('=')]))
            except ValueError:
                ret[item] = ''

        return ret

    def __updateResources(self):
        """
        Updates resources dictionary for this host
        """
        try:
            self.resources_dict  = dict([raw_res.split('=') for raw_res in self['resources'].split(',')])
        except ValueError:
            self.resources_dict = self.__slowProps('resources')
        except KeyError:
            pass


    def __updateProperties(self):
        """
        updates properties dictionary for this host
        """
        try:
            self.properties_dict = dict([raw_prop.split('=') for raw_prop in self['properties'].split(',')])
        except ValueError:
            self.properties_dict = self.__slowProps('properties')
        except KerError:
            pass


    def __updateFrames(self,force=False):
        """
        Update the list of frames that this host is running at this moment
        The list of frames is returned for conveniece, though one could also access them using the
        property 'frames'
        """
        if self.__frames and not force:
            logging.debug("Returning cached frames for host '%s'" % self.get('name',''))
            return self.__frames()

        logging.debug("Updating frames for host '%s'" % self.get('name',''))
        subjob_ids = ["%s.%s" % (sj['pid'],sj['id']) for sj in self.get('subjobs',[])]

        # if there are no subjobs, there are no frames
        if not subjob_ids:
            return []

        # update our jobs if we need to:
        if force or not self.__jobs:
            self.__updateJobs(force=force)

        self.__frames = []
        for job in self.__jobs:
            self.__frames += [frame for frame in job.frames if "%s.%s" % (frame['pid'],frame['subid']) in subjob_ids if frame.get('status','') == 'running']

        return self.__frames
        
        

    def __updateSubjobs(self,force=False):
        """
        Update the list of subjobs that this host is running at this moment
        The list of subjobs is returned for conveniece, though one could also access them using the
        property 'subjobs'
        """
        if self.__subjobs and not force:
            logging.debug("Returning cached subjobs for host '%s'" % self.get('name',''))
            return self.__subjobs
        
        logging.debug("Updating subjobs for host '%s'" % self.get('name',''))
        subjob_ids = ["%s.%s" % (sj['pid'],sj['id']) for sj in self.get('subjobs',[])]
        if not subjob_ids:
            return []

        # update our jobs if we need to:
        if force or not self.__jobs:
            self.__updateJobs(force)

        self.__subjobs = []
        for job in self.__jobs:
            self.__subjobs += [sj for sj in job.subjobs if "%s.%s" % (sj['pid'],sj['id']) in subjob_ids if sj.get('status','') == 'running']

        return self.__subjobs


    def __updateJobs(self,force=False):
        """
        Update the list of jobs that this host is running at this moment.
        The list of jobs is returned for conveniece, though one could also access them using the
        property 'jobs'
        """
        if self.__jobs and not force:
            logging.debug("Returning cached jobs for host '%s'" % self.get('name',''))
            return self.__jobs

        logging.debug("Updating jobs for host '%s'" % self.get('name','UNKNOWN'))
        self.__jobs = []
        for jid in set([sj['pid'] for sj in self.get('subjobs',[])]):
            try:
                if force:
                    logging.debug("Host '%s' requesting an update of its running jobs" % self.get('name','UNKNOWN'))
                    self.__job_list[jid].updateDetail()
                self.__jobs.append(self.__job_list[jid])
            except:
                logging.debug("Host could not get job %d.\n\t%s" % (jid,traceback.format_exc().replace("\n","\n\t")))
                continue

        return self.__jobs
    
        
    def hasDetail(self):
        """
        Accessor function for the private 'has_detail'
        returns the val of the var
        """
        return self.__has_detail


    def getConfig(self,force_update=False):
        """
        returns the current worker configuration as a dictionary.
        """
        if not self.__config or force_update:
            if self.get('state','') not in ('active','panic'):
                logging.info("Cannot connect to worker '%s' to get config.  Skipping." % self.get('name',self.get('address')))
            else:
                logging.info("Fetching worker config for '%s'" % self.get('name'))
                try:
                    self.__config = qb.workerconfig(self['address'])
                except KeyError:
                    pass
        return self.__config



