"""
 Licensed Materials - Property of Pipelinefx L.L.C.


 (C) COPYRIGHT Pipelinefx Limited Liability Corporation.
  All Rights Reserved.
                               
  US Government Users Restricted Rights - Use, duplication or
  disclosure restricted by GSA ADP Schedule Contract with 
  Pipelinefx L.L.C.

 $DateTime: 2013/07/08 14:29:16 $
 $Revision: #3 $
 $Change: 11473 $
 $File: //depot/main/qubeMobileServer/qubeMobile/model/qbUIHostList.py $

 Written by: Brian Knudson / PFX
             briank@pipelinefx.com
             2011-10-05

 Description:
  UIHostList class - Holds a list of a qbUIoHsts with sensible access
  to host methods.

"""

import os
import sys
import re
import time
import datetime
import logging

try:
    import qb
except ImportError:
    api_path = "/Applications/pfx/qube/api/python/"
    if api_path not in sys.path:
        sys.path.insert(0,api_path)
    import qb


from qbUIList import QbUIList
from qbUIHost import QbUIHost


"""
Qb UI HostList Class
"""
    


class QbUIHostList(QbUIList):
    """
    A host list model.
    Can be contructed:
      blank: which will get all hosts from the db
      list: will create an internal dict from the list of hosts
      dict: will copy the given dict to internal
    """
    def __init__(self, host_list=[],job_list=[]):
        logging.debug("Creating QbUIHostList")
        super(QbUIHostList,self).__init__(host_list)

        # this is not strictly necessary, though it maintains some similarity between jobs & hosts
        self.qb_hostinfo = qb.hostinfo

        # if we were passed a dict, we've already ini'd our parent, so we're done,
        #  otherwise, we've got some updatin' to do:
        if not isinstance(host_list,dict):
            self.updateHostList(host_list=host_list,job_list=job_list,clear=True)


    ###
    #
    #  Custom methods for this class:
    #
    def updateHostList(self,filters={},clear=False,host_list=None,job_list=None):
        """
        Populates internal dictionary with the return of an API call.
        Params:
          clear=False:  If clear is True, the internal dict is cleared before being populated
                        If clear is False (default), only the entries in the internal dict that
                         are newer than our internal timestamp are updated.
          hostList=None: If supplied, use this last rather than fetching one via the api.
          job_list=None: If supplied, pass the joblist through to each host
        Pre-condition: None
        Post-condition: internal dictionary has complete of partial new data
        Return value: None
        """
        if clear:
            # self._data.clear()
            self._data = {}
            self.last_updated = 0

        # get the time before the call, in the event the call takes forever to return
        start = now = int(time.time())

        if host_list:
            # if we were given a list of hosts, check to see if we need to
            # update our last_updated
            self.last_updated = max(h.get('lastupdate',0) for h in host_list.values())
        else:
            host_list = self.qb_hostinfo(filters=filters,subjobs=True)
            self.last_updated = now

        # updated the list with what we just got.  Dirty up the entry so we know it's new
        for h in host_list:
            self._data[h['macaddress']] = QbUIHost(host=h,job_list=job_list)

        end = time.time()

        logging.info("Updated %d hosts(s) in %.2fs" % (len(host_list),end-start))

        
    def getHostStats(self):
        """
        return a statistics dictionary for all host in the hostlist
        """
        logging.info("Getting host list stats")
        if not self._data:
            logging.info("No host list stats because host list is empty")
            return {}

        ret = {"active_workers":0,
               "total_workers":0,
               "active_slots":0,
               "total_slots":0}
        
        for mac,host in self._data.items():
            if host.state() not in ("active","running","idle"):
                continue
            ret["total_workers"] += 1
            used,total = map(int,host.resources_dict.get("host.processors","0/0").split('/'))
            ret["total_slots"] += total
            ret["active_slots"] += used
            if used > 0:
                ret["active_workers"] += 1

        return ret
