"""
 Licensed Materials - Property of Pipelinefx L.L.C.


 (C) COPYRIGHT Pipelinefx Limited Liability Corporation.
  All Rights Reserved.
                               
  US Government Users Restricted Rights - Use, duplication or
  disclosure restricted by GSA ADP Schedule Contract with 
  Pipelinefx L.L.C.

 $DateTime: 2013/07/08 14:29:16 $
 $Revision: #6 $
 $Change: 11473 $
 $File: //depot/main/qubeMobileServer/qubeMobile/model/qbUIJobList.py $

 Written by: Brian Knudson / PFX
             briank@pipelinefx.com
             2011-10-05

 Description:
  UIJobList class - Holds a list of a qbUIJobs with sensible access
  to job methods.

"""

import os
import sys
import re
import time
import datetime
import logging
from operator import itemgetter

try:
    import qb
except ImportError:
    for api_path in ("/Applications/pfx/qube/api/python/",
                     "/usr/local/pfx/qube/api/python/",
                     "C:\\Program Files\\pfx\\qube\\api\\python",
                     "C:\\Program Files (x86)\\pfx\\qube\\api\\python"):
        if api_path not in sys.path and os.path.exists(api_path):
            sys.path.insert(0,api_path)
            break
    import qb

from qbUIList import QbUIList
from qbUIJob import QbUIJob
from qbUILogList import QbUILogList


class QbUIJobList(QbUIList):
    """
    A job list model.
    Can be contructed:
      blank: which will get all hosts from the db
      list: will create an internal dict from the list of hosts
      dict: will copy the given dict to internal
    """
    def __init__(self, filters={}, job_list=[], limit=50, sql_user='qube_readonly', where=None):
        logging.debug("Creating QbUIJobList")

        self.querySQL = True
        try:
            import qb.query
            db = qb.query.mysqlConnect(user=sql_user)
            dbHostInfo = db.get_host_info()
            db.close()
        except:
            self.querySQL = False


        super(QbUIJobList,self).__init__(job_list)

        self.limit=limit
        self.last_updated = 0
        self.sql_user = sql_user
        self.__child_dict = {} # will hold a mapping of pgrp leaders & "children"

        if self.querySQL:
            self.qb_jobinfo = qb.query.jobinfo
        else:
            self.qb_jobinfo = qb.jobinfo

        # if a cient_logpath is given, pass it to the jobs (doing it here means
        # one stat for all jobs rather than one stat per job if it was done in the
        # job...  being in the job seems to make more OO sense, but.. too expensive)
        try:
            self.__mounted_logs = os.path.exists(qb.localconfig().get("client_logpath"))
            if self.__mounted_logs:
                logging.info("Use logfile location specified by client_ligpath: '%s'" % qb.localconfig().get("client_logpath"))
            else:
                logging.info("Fetching logfiles from supervisor")
        except:
            logging.info("Fetching logfiles from supervisor")
            self.__mounted_logs = False


        self.__log_list = QbUILogList(job_list=self)

        # NOTE: This is weird logic... designed to fit the class = eval(repr(class)) idea.
        #  Probably needs a re-think.
        # if we were passed a dict, we've already ini'd our parent, so we're done,
        #  otherwise, we've got some updatin' to do:
        #if not isinstance(job_list,dict):
        if not self._data:
            self.updateJobList(filters=filters,job_list=job_list,clear=True,where=where)



    ###
    #
    #  Convenience methods (not coming from base class):
    #
    def __str__(self):
        """
        str(object) -> string

        Return a nice string representation of the object.
        If the argument is a string, the return value is the same object.
        """
        keys = sorted(self._data.keys())
        start = end = 0
        if keys:
            start = keys[0]
            end = keys[-1]
        return "QbUIJobList. size: %d, id rage: %d-%d, last updated: %s" % (len(self._data),
                                                                            start,
                                                                            end,
                                                                            time.ctime(self.last_updated))

    def __del__(self):
        """
        Destructor that makes sure we explcitly clean up
        """
        logging.info("Joblist destructor...")
        self.clearAll()

    ###
    #
    #  Custom methods for this class:
    #
    def clearAll(self):
        """
        Explicitly calls the clearCache method of all jobs.  It appears that their
        destructor is not always called, therefore leaving cache files in tmp.  This
        prevents that.
        Also explicitly clears the job (_data) dict.
        """
        logging.info("Job list clearing all log cache files...")
        self.__log_list.clearAll()
        #self._data.clear() # maybe redundant?
        self._data = {}
        self.last_updated = 0


    def getLogList(self):
        """
        Returns the loglist model.  Mainly used for debugging
        """
        return self.__log_list
    

    def updateJobList(self,filters={},clear=False,job_list=None,where=None,limit=None):
        """
        Populates internal dictionary with the return of an API call.
        Params:
          clear=False:  If clear is True, the internal dict is cleared before being populated
                        If clear is False (default), only the entries in the internal dict that
                         are newer than our internal timestamp are updated.
          jobList=None: If supplied, use this last rather than fetching one via the api.
        Pre-condition: None
        Post-condition: internal dictionary has complete of partial new data
        Return value: None
        """
        start = time.time()
        if clear:
            self.clearAll()

        # get the time before the call, in the event the call takes forever to return
        now = int(time.time())

        if limit:
            self.limit=limit

        if job_list:
            # if we were given a list of jobs, check to see if we need to
            # update our last_updated
            self.last_updated = max(j.get('lastupdate',0) for j in job_list.values())
        else:
            if self.querySQL:
                logging.info("Joblist updating, using direct SQL queries.")
                job_list = self.qb_jobinfo(filters=filters,updatedAfter=self.last_updated,
                                           limit=self.limit,sortOrder='DESC',sqluser=self.sql_user,where=where)
            else:
                logging.info("Joblist updating through API (no direct SQL).")
                job_list = sorted(self.qb_jobinfo(filters=filters),key=itemgetter('id'),reverse=True)[:self.limit]
                if where:
                    logging.warn("Server side search attempted on a system without direct mysql access. Search failed")
            # Save last_updated time as now - 1s.  This covers the case when a job changes state more than
            # onces in one second, which can happen with job kills, for example.
            self.last_updated = now - 1 

        for j in job_list:
            self._data[j['id']] = QbUIJob(j,log_list=self.__log_list,sql_user=self.sql_user,query_sql=self.querySQL)
            # check to see if this job is a "child" of some other pgrp leader
            # if so, save it for later - this way we know the whole job list will
            # be populated before we eat a dick
            parent_id = self._data[j['id']].get('pgrp')
            if parent_id != j['id']:
                if not self.__child_dict.has_key(parent_id):
                    self.__child_dict[parent_id] = set()
                self.__child_dict[parent_id].add(j['id'])

        # now that the entire joblist is populated, go back and update children
        for parent,children in self.__child_dict.items():
            try:
                self._data[parent].children.update(children)
            except KeyError:
                logging.debug("Missing parent %d for child %s in joblist")

        end = time.time()

        logging.info("Updated %d job(s) in %.2fs" % (len(job_list),end-start))

        
    def getJobStats(self):
        """
        return a statistics dictionary for all job in the joblist
        dict will contain pending/blocked/running jobs/frames
        """
        logging.info("Getting job list stats")
        if not self._data:
            logging.info("No job list stats because job list is empty")
            return {}

        ret = {"pending_jobs":0,
               "running_jobs":0,
               "blocked_jobs":0,
               "pending_frames":0,
               "running_frames":0,
               "blocked_frames":0}

        for id,job in self._data.items():
            status = job.get("status",None)
            if status not in ("running","pending","blocked"):
                continue
            if status == "pending":
                ret["pending_jobs"] += 1
            elif status == "running":
                ret["running_jobs"] += 1
            else:
                ret["blocked_jobs"] += 1

            try:
                ret["pending_frames"] += job.getAttribtue("pending")
                ret["running_frames"] += job.getAttribtue("running")
                ret["blocked_frames"] += job.getAttribtue("blocked")
            except:
                pass

        return ret
