"""
 Licensed Materials - Property of Pipelinefx L.L.C.


 (C) COPYRIGHT Pipelinefx Limited Liability Corporation.
  All Rights Reserved.
                               
  US Government Users Restricted Rights - Use, duplication or
  disclosure restricted by GSA ADP Schedule Contract with 
  Pipelinefx L.L.C.

 $DateTime: 2013/07/08 14:29:16 $
 $Revision: #3 $
 $Change: 11473 $
 $File: //depot/main/qubeMobileServer/qubeMobile/model/qbUIList.py $

 Written by: Brian Knudson / PFX
             briank@pipelinefx.com
             2011-10-05

 Description:
  Base class for all UI*List classes
"""

import time
import logging

class QbUIList(object):
    def __init__(self,data={}):
        logging.debug("Creating QbUIList")
        self._data        = data       # holds out list of jobs, hosts, etc
        self.last_updated = None       # the time of the last update
        self.dirty        = []         # list of jobs/hosts that are "dirty" - have just updated.  NOTE: may not be necessary!


    ###
    #
    #  Convenience methods:
    #
    def __repr__(self):
        """
        repr(object) -> string

        Return the canonical string representation of the object.
        For most object types, eval(repr(object)) == object.
        """
        return "%s(%s)" % (self.__class__.__name__,self._data.values())

    def __str__(self):
        """
        str(object) -> string

        Return a nice string representation of the object.
        If the argument is a string, the return value is the same object.
        """
        return "Instance of %s. size: %d, last updated: %s" % (self.__class__.__name__,
                                                               len(self._data),
                                                               time.ctime(self.last_updated))

    def __len__(self):
        """
        Implementation of len(self)
        """
        return len(self._data)

    def __getitem__(self,key):
        """
        Implementation of self[key] on the fetching side.
        self[key] will be a qb.Job.
        """
        return self._data[key]

    def __setitem__(self,key,value):
        """
        Implementation of self[key] on the setting side.
        self[key] will be a new value or a replacement value.
        """
        self._data[key] = value

    def __delitem__(self,key):
        """
        x.__delitem__(y) <==> del x[y]
        """
        del(self._data[key])

    def __iter__(self):
        """
        Implementation of an iterator for the job list.
        """
        yield self._data.keys()

    def __contains__(self,a):
        """
        D.__contains__(k) -> True if D has a key k, else False
        """
        return self._data.__contains__(a)

    ###
    #
    #  Methods to mymic python built-in types:
    #
    def keys(self):
        """
        J.keys() -> list of J's keys
        """
        return self._data.keys()


    def has_key(self,k):
        """
        J.has_key(k) -> True if J has a key k, else False
        """
        return self._data.has_key(k)

    def values(self):
        """
        J.values() -> list of J's values
        """
        return self._data.values()

    def items(self):
        """
        J.items() -> list of J's (key, value) pairs, as 2-tuples.
        """
        return self._data.items()

    def get(self,*args):
        """
        J.get(k[,d]) -> J[k] if k in J, else d.  d defaults to None.
        """
        return self._data.get(*args)

    def clear(self):
        """
        J.clear() -> None.  Remove all items from J.
        """
        self._data.clear()
        self.dirty = []  # no more ditry rows
        self.last_updated = 0  # there's no data, so don't try to just updated anymore

    def setdefault(self,*args):
        """
        If self[key] doesn't exist, it gets set to 'value' & self[key] is returned.
        If self[key] is set, 'value' is discarded & self[key] is returned.
        """
        return self._data.setdefault(*args)

    def iterkeys(self):
        """
        J.iteritems() -> an iterator over the keys of J
        """
        yield self._data.keys()

    def iteritems(self):
        """
        J.iteritems() -> an iterator over the (key, value) items of J
        """
        yield self._data.items()

    def pop(self,*args):
        """
        J.pop(k[,d]) -> v, remove specified key and return the corresponding value.
        If key is not found, d is returned if given, otherwise KeyError is raised
        """
        return self._data.pop(*args)

    def popitem(self,*args):
        """
        J.popitem() -> (k, v), remove and return some (key, value) pair as a
        2-tuple; but raise KeyError if J is empty.
        """
        return self._data.popitems(*args)

    def copy(self):
        """
        J.copy() -> a shallow copy of J
        """
        return self._data.copy()

    def update(self,*a,**b):
        """
        J.update(E, **F) -> None.  Update J from dict/iterable E and F.
        If E has a .keys() method, does:     for k in E: J[k] = E[k]
        If E lacks .keys() method, does:     for (k, v) in E: J[k] = v
        In either case, this is followed by: for k in F: J[k] = F[k]
        """
        self._data.update(*a,**b)
        
    def findHash(self,**kwargs):
        """
        Given a key,value pair, findItem will return the hash of the first
        item in the dict that has a matching key/val.  This assumes dicts are
        in the form {k: {k:v, k:v}}

        Example:
        myHostList.findHash(name='myHostName')
        """
        try:
            sk,sv = kwargs.items()[0]
        except IndexError:
            return None

        if not isinstance(self._data,dict):
            return None

        for k,item in self._data.items():
            for key,val in item.items():
                if key == sk and val == sv:
                    return k
        
