"""
 Licensed Materials - Property of Pipelinefx L.L.C.


 (C) COPYRIGHT Pipelinefx Limited Liability Corporation.
  All Rights Reserved.
                               
  US Government Users Restricted Rights - Use, duplication or
  disclosure restricted by GSA ADP Schedule Contract with 
  Pipelinefx L.L.C.

 $DateTime: 2013/07/08 14:29:16 $
 $Revision: #4 $
 $Change: 11473 $
 $File: //depot/main/qubeMobileServer/qubeMobile/model/qbUILog.py $

 Written by: Brian Knudson / PFX
             briank@pipelinefx.com
             2011-10-05

 Description:
  UILog class - holds log data on a per job bases with accessor methods
  of retrieving that data on a per job, subjob, or frame level

"""

import os
import re
import sys
import mmap
import logging
import tempfile
import glob
import traceback
from operator import itemgetter

try:
    import qb
except ImportError:
    for api_path in ("/Applications/pfx/qube/api/python/",
                     "/usr/local/pfx/qube/api/python/",
                     "C:\\Program Files\\pfx\\qube\\api\\python",
                     "C:\\Program Files (x86)\\pfx\\qube\\api\\python"):
        if api_path not in sys.path and os.path.exists(api_path):
            sys.path.insert(0,api_path)
            break
    import qb

class QbUILog(object):
    def __init__(self,job_id,job_list,mounted_logs=False,sql_user='qube_readonly'):
        self.__job_id = job_id
        self.__job_list = job_list
        self.__mounted_logs = mounted_logs

        self.__local_temp_dir = None   # holds the path to the local tmp dir
        self.__local_stdout_path = []  # holds the path(s) to the local stdout files - one path per subjob
        self.__local_stderr_path = []  # holds the path(s) to the local stderr files - one path per subjob
        self.__has_logs = False        # Do we have logs or are we empty?
        self.sql_user = sql_user       # user for direct sql queries


    def __del__(self):
        """
        destructor that removes any temp cache files that may exist
        """
        logging.debug("Running destructor for log '%d'" % self.__job_id)
        self.clearCache()


    def clearCache(self):
        """
        Removes cache files on the local FS if they exist
        """
        # Could use: shutil.rmtree(self.__local_temp_dir)... not sure why I'm not doing that.  hmmm...
        if not self.__local_temp_dir:
            return
        #shutil.rmtree(self.__local_temp_dir)
        if os.path.exists(self.__local_temp_dir):
            files = os.listdir(self.__local_temp_dir)
            for f in files:
                if not f:
                    continue
                path = os.path.join(self.__local_temp_dir, f)
                if not os.path.isfile(path):
                    continue
                logging.debug("removing cache file '%s'" % f)
                try:
                    os.unlink(path)
                except Exception,e:
                    logging.warn("Could not remove cache file '%s' : %s" % (path,str(e)))
            logging.info("removing cache dir '%s'" % self.__local_temp_dir)
            os.rmdir(self.__local_temp_dir)



    def __updateLogCacheFiles(self):
        """
        updates the local log files with data requested from the supevisor.
        This function is only called with updateDetail is called
        """
        self.__local_stdout_path = []
        self.__local_stderr_path = []
        logging.info("Retrieving job logs for job '%d'" % self.__job_id)

        # if we have mounted logs (which means client_logpath is defined), then don't use local tmp.
        if self.__mounted_logs:
            job_containing_dir = self.__job_id / 1000 * 1000
            try:
                job_log_loc = os.path.join(qb.localconfig().get('client_logpath'),"job",str(job_containing_dir),str(self.__job_id))
            except AttributeError:
                msg =  "Attempted to find logpath for job '%d', but client_logpath seems to no longer be defined." % self.__job_id
                msg += " Reverting to remote logs."
                logging.warn(msg)
                job_log_loc = '?|?|?|?|?|?|?|' # non-sensical, obvious missing path to force the next test to fail
            if os.path.exists(job_log_loc):
                logging.debug("Using mounted path '%s' for job '%d' logs" % (job_log_loc,self.__job_id))
                self.__local_stdout_path = glob.glob(os.path.join(job_log_loc,"*.out"))
                self.__local_stderr_path = glob.glob(os.path.join(job_log_loc,"*.err"))
                self.__has_logs = True
                return


        if not self.__local_temp_dir or not os.path.exists(self.__local_temp_dir):
            self.__local_temp_dir = tempfile.mkdtemp(prefix='%s-%d' % (self.__class__.__name__,self.__job_id))

        # Be sure the list of subjobs is sorted by id in this call:
        # subjobs = ('%d.%d' % (self.__job_id,sj['id']) for sj in sorted(self['subjobs'],key=itemgetter('id')))

        for sj in sorted(('%d.%d' % (self.__job_id,sj['id']) for sj in self.__job_list[self.__job_id].subjobs)):
            logging.debug("Working on log files for subjob %s" % sj)
            for mode in "out","err":
                cache_file_path = os.path.join(self.__local_temp_dir,"%s.%s" % (sj.replace('.','_'),mode))
                # NOTE: Need to be able to force write for update
                if os.path.exists(cache_file_path):
                    logging.debug("skipping cache file '%s' because it already exists" % (cache_file_path))
                    if mode == "out":
                        self.__local_stdout_path.append(cache_file_path)
                    else:
                        self.__local_stderr_path.append(cache_file_path)
                else:
                    logging.debug("Writing cache file '%s' for job '%s' logs" % (cache_file_path,sj))
                    if mode == "out":
                        open(cache_file_path,'w').write(qb.stdout(sj)[0].get('data',"NO DATA"))
                        self.__local_stdout_path.append(cache_file_path)
                    else:
                        open(cache_file_path,'w').write(qb.stderr(sj)[0].get('data',"NO DATA"))
                        self.__local_stderr_path.append(cache_file_path)


        # Make sure we have no duplicate paths:
        seen = set()
        self.__local_stdout_path = [x for x in self.__local_stdout_path if x not in seen and not seen.add(x)]
        self.__local_stderr_path = [x for x in self.__local_stderr_path if x not in seen and not seen.add(x)]

        self.__has_logs = True


    def updateLogs(self,force=False):
        """
        If logs haven't already been fetched, go get 'em.
        If 'force' is set to true, go get 'em regardless.
        """
        if not self.__has_logs or force:
            self.__updateLogCacheFiles()

        return self.getJobLogs()


    def getJobLogs(self):
        """
        Get logs for this job.
        Return value is a tuple of lists, return[0] is stdout, return[1] is stderr.
        Each item in the stdout/err list is a subjob log dictionary, as returned
        by qb.stdout/err(), which has the form of {jobid:'',subid:'',data''} where
        'data' is the actual log.
        """
        logging.info("Getting logs for job '%s'" % self.__job_id)
        if not self.__has_logs:
            self.__updateLogCacheFiles()

        stdout = []
        stderr = []
        p = re.compile("(\d+)_(\d+).(out|err)")
        for log_file in (self.__local_stdout_path,self.__local_stderr_path):
            log = {True:stdout,False:stderr}[log_file == self.__local_stdout_path]
            for f in log_file:
                try:
                    jobid,subid = re.search(p,f).group(1,2)
                except IndexError:
                    logging.info("Couldn't find one of the subjobs logs for job '%d'" % self.__job_id)
                    continue
                # This is to mymic the result of qb.stdout/err:
                log.append({'jobid':int(jobid),'subid':int(subid),'data':open(f).read()})

        return stdout,stderr


    def getSubjobLogs(self,subid):
        """
        Return the log for the given subjob_id.
        subjob_id should be given in the form of job_id.subjob_id
        return val is (stdout,stderr)
        """
        if not isinstance(subid,int):
            subid = int(subid)

        logging.debug("Getting logs for subjob '%s.%s'." % (self.__job_id,subid))
        ret = ['','']

        # make sure we know all of our subjob ids
        subjobs = sorted(('%d.%d' % (self.__job_id,sj['id']) for sj in self.__job_list[self.__job_id].subjobs))

        # make sure the requested subjob id exists for this [job's] log
        if subid not in (int(sj.split('.')[-1]) for sj in subjobs):
            logging.warn("Requested subjob id '%s' is not a known subjob for this job '%s'" % (subid,self.__job_id))
            return ret

        # make sure we have log data for this job
        if not self.__has_logs:
            self.__updateLogCacheFiles()

        stdout_path = self.__local_stdout_path[subid]
        stderr_path = self.__local_stderr_path[subid]

        # If the job hasn't started, it's possible we don't have a log path yet (unlikely)
        if not stdout_path and not stderr_path:
            logging.info("Paths not known for subjob '%s.%s'" % (self.__job_id,subid))
            return ret

        # even more likely is that the paths to the logfiles are not yet valid
        # (especially when client_logpath is defined and the job has not yet started)
        if os.path.getsize(stdout_path) == 0 and os.path.getsize(stderr_path) == 0:
            logging.info("Logs for subjob '%s.%s' are blank." % (self.__job_id,subid))
            return ret

        try:
            stdout = open(stdout_path).read()
        except IOError:
            stdout = None

        try:
            stderr = open(stderr_path).read()
        except IOError:
            stderr = None

        return stdout,stderr


    def __mmap_rfind(self,my_map,search,start=-1,end=-1):
        """
        Does what mmap.rfind would do if python2.5 had mmap.rfind... which it doesn't
        my_map: the map to use
        search: string to search for
        start: place to start looking
        end: place to stop looking (should be < start)
        """
        if start < 0:
            start = 0
        if end < 0:
            end = my_map.size()-1 # -1 to get to the last actual char

        for i in range(end,start-1,-1): # start-1 because range will not include the last number
            if my_map[i] == search:
                return i

        return -1


    def frameLogHelper(self,map,start,end,lock_end=False):
        """
        Takes a start and end position in a given mmap, adjusts the start and end
        back to find the most recent \n before their position (if it exists), then
        returns the string from the map that starts at the adjusted start and ends
        at the adjusted end.  If 'lock_end' is set, the end is not adjusted (good
        for strings that touch the end of the log file)
        """
        try:
            line_start = max(0,map.rfind('\n',0,start))
        except AttributeError:
            line_start = max(0,self.__mmap_rfind(my_map=map,search='\n',start=0,end=start))
        if lock_end:
            try:
                line_end = max(0,map.rfind('\n',0,end))
            except AttributeError:
                line_end = max(0,self.__mmap_rfind(my_map=map,search='\n',start=0,end=end))
        else:
            line_end = end
        logging.debug("frame log start/stop adjusted to %d/%d." % (line_start,line_end))

        return map[line_start:line_end]


    def getFrameLogs(self,frame_id):
        """
        Return the log for this one frame
        return val is (stdout,stderr)
        """
        ret = ['','']

        frame_name = self.__job_list[self.__job_id].frames[frame_id].get('name') # -1 to adjust for 1 based ids. ::grumble::
        subid = self.__job_list[self.__job_id].frames[frame_id].get('subid')
        jobframe =   '%d:%s' % (self.__job_id,self.__job_list[self.__job_id].frames[frame_id].get('name'))

        logging.debug("Frame name determined to be: %s" % frame_name)
        logging.debug("This frame has a subid of: %s" % subid)
        logging.debug("The jobframe for this frame: %s" % jobframe)

        for i,sop in enumerate(self.__local_stdout_path):
            logging.debug("stdout path %d: %s" % (i,sop))
        for i,sop in enumerate(self.__local_stderr_path):
            logging.debug("stderr path %d: %s" % (i,sop))


        # make sure we have log data for this job
        if not self.__has_logs:
            self.__updateLogCacheFiles()

        stdout_path = self.__local_stdout_path[subid]
        stderr_path = self.__local_stderr_path[subid]

        if not stdout_path and not stderr_path:
            return ret
        if os.path.getsize(stdout_path) == 0 and os.path.getsize(stderr_path) == 0:
            return ret

        # If a file is empty, you can't mmap it, so in the case where stdout & stderr are combined,
        # (or in any other case when you would have a log of 0 size)put some dummy data in one or the other.
        if os.stat(stdout_path).st_size > 0:
            out_fp = open(stdout_path)
        else:
            logging.debug("Making a dummy file for stdout frame log for '%d/%d' as it has a 0 file size." % (self.__job_id,frame_id))
            out_fp = tempfile.TemporaryFile(dir=os.path.dirname(stdout_path))
            out_fp.write('NO DATA')
            out_fp.flush()

        if os.stat(stderr_path).st_size > 0:
            err_fp = open(stderr_path)
        else:
            logging.debug("Making a dummy file for stderr frame log for '%d/%d' as it has a 0 file size." % (self.__job_id,frame_id))
            err_fp = tempfile.TemporaryFile(dir=os.path.dirname(stderr_path))
            err_fp.write('NO DATA')
            err_fp.flush()

        logging.debug("Using stdout log '%s' for frame '%d/%s'" % (stdout_path,self.__job_id,frame_id))
        logging.debug("Using stderr log '%s' for frame '%d/%s'" % (stderr_path,self.__job_id,frame_id))

        out_map = mmap.mmap(out_fp.fileno(),length=0,access=mmap.ACCESS_READ)
        err_map = mmap.mmap(err_fp.fileno(),length=0,access=mmap.ACCESS_READ)
            
        iter_pat = re.compile('got work: \d+:(\S+)')
        stdout_iter = re.finditer(iter_pat, out_map)
        stderr_iter = re.finditer(iter_pat, err_map)

        # self_pat = re.compile("%s(-\S+)?$" % jobframe)
        self_pat = re.compile("%s$" % jobframe)

        for count,iterator in enumerate((stdout_iter,stderr_iter)):
            start = end = -1
            found = False
            log_data = []
            for i in iterator:
                logging.debug("Parsing log, checking %s" % str(i.span()))
                if found:
                    end = i.span()[0]
                    found = False

                    logging.debug("%s frame log start/stop for frame '%s' of job '%d.%d': %d/%d" % ({0:'stdout',1:'stderr'}[count],
                                                                                                    frame_name,
                                                                                                    self.__job_id,
                                                                                                    subid,
                                                                                                    start,
                                                                                                    end))
                    log_data.append(self.frameLogHelper(map=(out_map,err_map)[count],start=start,end=end))

                    # Note, you could stop parsing at this point, but then you eliminate the possiblity
                    # of seeing any retries.


                # This would work (and be prettier) if we could use python2.6  in 2.5, however,
                # you can't seem to do string operations on an mmap, so...
                #if re.search(self_pat,(out_map,err_map)[count][slice(*i.span())]):
                (out_map,err_map)[count].seek(i.span()[0])
                search_string = (out_map,err_map)[count].read(i.span()[-1] - i.span()[0])
                if re.search(self_pat,search_string):
                    logging.debug("found the logs in %s" % str(i.span()))
                    start = i.span()[0]
                    found = True

            # The last frame log will find the start position on the last iteration of the
            # for i in iter loop, so it will NOT loop through to find the end position.  In that
            # case, we'll mark the end as the end of the file & be done with it.  You *could*
            # only use this end case (with a break in the "if found" section above), but you
            # would not catch retries. That said, retries from another subjob will not be found
            # even using this method so... yeah..  50% is better than 0%.
            logging.debug("start: %d, end: %d" % (start,end))
            if start > 0 and end < 0:
                logging.debug("%s frame log start/stop for frame '%s' of job '%d.%d': %d/%d" % ({0:'stdout',1:'stderr'}[count],
                                                                                                frame_name,
                                                                                                self.__job_id,
                                                                                                subid,
                                                                                                start,
                                                                                                end))
                log_data.append(self.frameLogHelper(map=(out_map,err_map)[count],start=start,end=end,lock_end=True))

            ret[count] = ("\n\n%s\n\n" % (" RETRY ".center(40,'-'))).join(log_data)
                    


        out_map.close()
        err_map.close()
        out_fp.close()
        err_fp.close()
        return ret


    def getSubjobLogPath(self,subid):
        """
        Returns the path to the log file for the requested subjob.
        return is (stdout_path, stderr_path)
        """
        try:
            return self.__local_stdout_path[subid],self.__local_stderr_path[subid]
        except IndexError:
            self.__updateLogCacheFiles()
            return self.__local_stdout_path[subid],self.__local_stderr_path[subid]
