"""
 Licensed Materials - Property of Pipelinefx L.L.C.


 (C) COPYRIGHT Pipelinefx Limited Liability Corporation.
  All Rights Reserved.
                               
  US Government Users Restricted Rights - Use, duplication or
  disclosure restricted by GSA ADP Schedule Contract with 
  Pipelinefx L.L.C.


 QubeView Subjob module

 Name: qbUISubjob.py

 $DateTime: 2013/07/08 14:29:16 $
 $Revision: #3 $
 $Change: 11473 $
 $File: //depot/main/qubeMobileServer/qubeMobile/model/qbUISubjob.py $

 Written by: Brian Knudson / PFX
             briank@pipelinefx.com
             2011-12-05

 Description:
  Subjob module, subclassed off of qb.Subjob with some extra functions to
  access logs & human redable time.

"""

import os
import sys
import re
import time
import datetime
import mmap
import logging

try:
    import qb
except ImportError:
    api_path = "/Applications/pfx/qube/api/python/"
    if api_path not in sys.path:
        sys.path.insert(0,api_path)
    import qb

from qbUIBase import QbUIBase

class QbUISubjob(qb.Subjob,QbUIBase):
    """
    A subjob model.
    Construct with:
     subjob: a qb.Work object
     out_path: path to the owning subjob's stdout cache file
     err_path: path to the owning subjob's stderr cache file
    Note: Both outfp & errfp can be None, but they will never be updated in that case
    """
    attribute_map = [{"display":"Started",  "attribute":"getReadableTimeStart()",    "name":"started"},
                     {"display":"Completed","attribute":"getReadableTimeComplete()", "name":"completed"},
                     {"display":"Elapsed",  "attribute":"getReadableElapsed()",      "name":"elapsed"},
                     {"display":"CPU-minutes", "attribute":"getTimeCumulative()",    "name":"time_cumulative"},
                     {"display":"id"},
                     {"display":"pid"},
                     {"display":"host"},
                     {"display":"count"},
                     {"display":"retry"},
                     {"display":"status"},
                     {"display":"slots"},
                     {"display":"lastupdate"},
                     {"display":"timesubmit"},
                     {"display":"timestart"},
                     {"display":"timecomplete"},
                     ]

    def __init__(self,subjob={},log_list={}):
        super(QbUISubjob,self).__init__(subjob)
        self.__log_list = log_list
        
    def getLogs(self,update=False):
        """
        Return the log for this one subjob
        return val is (stdout,stderr)
        """
        logging.debug("Fetching logs for subjob '%s.%s'." % (self.get('pid','--'),self.get('id','--')))
        try:
            return self.__log_list[self['pid']].getSubjobLogs(self['id'])
        except KeyError:
            return {}
    
    def getTimeCumulative(self):
        try:
            return str(datetime.timedelta(seconds=self.get('timecumulative','--')))
        except TypeError:
            return '--'

    def getReadableTimeStart(self):
        try:
            return { True:time.strftime("%X",time.localtime(self["timestart"])), False:"--" }[ self["timestart"] > qb.QB_TIME_EPOCH_OFFSET ]
        except:
            return "--"

    def getReadableTimeComplete(self):
        try:
            return { True:time.strftime("%X",time.localtime(self["timecomplete"])), False:"--" }[ self["timecomplete"] > qb.QB_TIME_EPOCH_OFFSET ]
        except:
            return "--"

    def getReadableElapsed(self):
        """
        returns the time this subjob has been running in a human
        readable string.
        """
        now = int(time.time())
        start_epoch    = {True:self['timestart'],    False:now}[self['timestart'] > qb.QB_TIME_EPOCH_OFFSET ]
        complete_epoch = {True:self['timecomplete'], False:now}[self['timecomplete'] > qb.QB_TIME_EPOCH_OFFSET ]
        elapsed_epoch = complete_epoch - start_epoch

        return {True:str(datetime.timedelta(seconds=elapsed_epoch)), False:"--" }[elapsed_epoch > 0]


