#!/usr/bin/env python

"""
#######################################################################
#
# Licensed Materials - Property of Pipelinefx L.L.C.
#
#
# (C) COPYRIGHT Pipelinefx Limited Liability Corporation.
#  All Rights Reserved.
#                               
#  US Government Users Restricted Rights - Use, duplication or
#  disclosure restricted by GSA ADP Schedule Contract with 
#  Pipelinefx L.L.C.
#
# $DateTime: 2012/12/17 15:21:39 $
# $Revision: #6 $
# $Change: 10481 $
# $File: //depot/main/qubeMobileServer/qubeMobileServerUserAdmin.py $
#
# Written by: Brian Knudson / PFX
#             briank@pipelinefx.com
#             2012-06-04
#
# Description:
#  User administration tool for Qube MobileServer.
#
# Notes:
#  The password database that is created is a simple, pickled file
#  that contains hashed passwords used for one-way password challenges.
#  While the data is obfuscated, it should not be considered secure.
#  That said, being that the stored data is a hash, even if one were
#  to pull the data from the database, it should not comprimise
#  security.
#
#######################################################################
"""

import sys
import getpass
import os
from optparse import OptionParser, Option
from config.qubeMobileServerSettings import QubeMobileServerSettings
from qubeMobile.model.qbUIMobileServerUser import QubeMobileServerUserDB


def askPass():
    """
    asks the user for a password, confirms it, makes sure they match,
    returns it in cleartext.
    """
    pw1 = getpass.getpass()
    pw2 = getpass.getpass(prompt='Confirm:')

    if pw1 == pw2:
        return pw1
    print >>sys.stderr, "Passwords do not match."
    return False

def askUser(question):
    """
    asks the user for a password, confirms it, makes sure they match,
    returns it in cleartext.
    """
    return raw_input(question)


def printUsers(password_file,extra=False):
    """
    pretty print for the user list
    """
    pw_db = QubeMobileServerUserDB(password_file=os.path.abspath(password_file))

    user_header = "User"
    passwd_header = "Encrypted password"

    user_chars = len(user_header)
    passwd_chars = len(passwd_header)
    for user,passwd in pw_db.users:
        user_chars = max(user_chars,len(user))
        passwd_chars = max(passwd_chars,len(passwd))

    border = "+-%s-+-%s-+" % ('-'.rjust(user_chars,'-'),'-'.rjust(passwd_chars,'-'))

    print >> sys.stderr,"\n" + border
    print >> sys.stderr,"| %s | %s |" % (user_header.ljust(user_chars), passwd_header.ljust(passwd_chars))
    print >> sys.stderr,border
    for user,passwd in pw_db.users:
        print >> sys.stderr,"| %s | %s |" % (user.ljust(user_chars), passwd.ljust(passwd_chars))
    print >> sys.stderr,border

    if extra:
        print >>sys.stderr, "%d users in database\n" % (pw_db.getSize())
    

def main(user_add=False,
         user_drop=False,
         password_file=None,
         hashed_password=False,
         list_users=False):
    pw_db = QubeMobileServerUserDB(password_file=os.path.abspath(password_file))
    db_size = pw_db.getSize()

    if list_users:
        printUsers(password_file,extra=True)
        sys.exit(0)
        
    if user_drop:
        users = []
        if user_drop[0] == "--no-arg-given--":
            user = askUser("Username to drop: ")
            users.append(user)
        else:
            users = user_drop
        for user in users:
            pw_db.dropUser(user=user)

    if user_add:
        if user_add[0] == "--no-arg-given--":
            user = askUser("Username to add: ")
        else:
            user = user_add[0]

        if len(user_add) > 1:
            password = user_add[1]
        else:
            password = askPass()
            if not password:
                print >> sys.stderr, "No password given. Exiting"
                sys.exit(1)


    if user_add:
        pw_db.addUser(user=user,password=password,hashed=hashed_password)
        if not pw_db.hasUser(user=user):
            print >> sys.stderr, "An error occured.  User was not added."
            sys.exit(1)
    printUsers(password_file)

    new_db_size = pw_db.getSize()
    print >>sys.stderr, "%d users in database. (%d user(s) %s)\n" % (new_db_size,
                                                                     abs(db_size - new_db_size),
                                                                     {True:'added',False:'removed'}[(db_size - new_db_size) <= 0])



def vararg_callback(option, opt_str, value, parser):
    assert value is None
    value = []

    def floatable(str):
        try:
            float(str)
            return True
        except ValueError:
            return False

    for arg in parser.rargs:
        # stop on --foo like options
        if arg[:2] == "--" and len(arg) > 2:
            break
        # stop on -a, but not on -3 or -3.0
        if arg[:1] == "-" and len(arg) > 1 and not floatable(arg):
            break
        value.append(arg)

    if not value:
        value = ['--no-arg-given--']

    del parser.rargs[:len(value)]
    setattr(parser.values, option.dest, value)

     
if __name__ == "__main__":
    launch_cwd = os.path.abspath(os.getcwd())
    script_location = os.path.abspath(os.path.dirname(sys.argv[0]))
    if script_location != launch_cwd:
        os.chdir(script_location)

    settings = QubeMobileServerSettings(os.path.join(script_location,"config","qbms.conf"))

    default_settings_file = os.path.abspath(settings.conf_file),
    default_password_file = settings.get("password_file",os.path.join(launch_cwd,".passwd"))


    #-------------------------------------------
    #default_password_file = qubeMobileServerConfig.password_file
    
    passwd_help  = "Location of the server-created password file.  "
    passwd_help += "Note this should absolutely NOT be a system-managed password file. "
    passwd_help += "Default: '%s'." % default_password_file

    usage = "%prog [options] username [password]"
    p = OptionParser(usage=usage)
    p.add_option("",  "--settings-file",
                 dest="settings_file",
                 default=default_settings_file,
                 action="store",
                 type="string",
                 help="Location of the qbms.conf file. Default: %s" % default_settings_file)
    p.add_option("", "--password-file",
                 dest="password_file",
                 default=default_password_file,
                 action="store",
                 type="string",
                 help=passwd_help)
    p.add_option("-a",  "--add",
                 dest="user_add",
                 action="callback",
                 callback=vararg_callback,
                 help="Add a user.  You can provide both or neither the username and/or password.")
    p.add_option("-d",  "--drop",
                 dest="user_drop",
                 action="callback",
                 callback=vararg_callback,
                 help="Drop a user.  You can optionally provide username or names, separated by spaces.")
    p.add_option("",  "--hashed",
                 dest="hashed",
                 default=False,
                 action="store_true",
                 help="The given password is already hashed")
    p.add_option("",  "--list",
                 dest="list_users",
                 default=False,
                 action="store_true",
                 help="List all users in the database then exit.")

    options,args = p.parse_args(sys.argv)


    if options.settings_file != default_settings_file:
        settings = QubeMobileServerSettings(options.settings_file)

    settings.update(vars(options))

    main(user_add=options.user_add,
         user_drop=options.user_drop,
         password_file=settings.get("password_file"),
         hashed_password=options.hashed,
         list_users=options.list_users)
